#include <stdlib.h>
#include <stdio.h>

#include "wishGui.hxx"
#include "cmumovie.h"


void wishGui::Initialize(void)
{
	WishIntfc::Initialize();
}

int wishGui::Start(char *tcl_fname)
{
	if (WishIntfc::Start(tcl_fname)==ERROR) return ERROR;
	return 0;
}

int wishGui::End(void)
{
	if (WishIntfc::End() == ERROR)
		return ERROR;
	return 0;
}

int wishGui::ProcessWishReq()
{
	char c;
	while (read(fromWish, &c, 1) > 0 ) {
		putc(c,stderr);
	}
	
	return NOERROR;
}

int global_cmuview_file=0;
char global_cmuview_filename[256];

wishGui wg;
static int width,height;
static FILE *fp;

int cmumovie_poll_wish()
{
	wg.ProcessWishReq();
}

int cmumovie_open(int delay)
{
	char buf[100];
	if (global_cmuview_file==0) {
		wg.SetPath("/usr/bin");
		if (wg.Start("cmuview2.tcl")==ERROR) {
			fprintf(stderr,"wg.Start failed!\n");
			return ERROR;
		}
		sprintf(buf,"cmumovie 1 0 %i\n",delay);
		return cmumovie_command(buf);
	} else {
		fp=fopen(global_cmuview_filename,"w");
		if (!fp) {
			fprintf(stderr,"Couldn't open file %s\n",global_cmuview_filename);
			exit(1);
		}
		fprintf(fp,"cmumovie 1 0 %i\n",delay);
		
	}

	return 0;
}

int cmumovie_close()
{
	return wg.End();
}

int cmumovie_command(char *cmd)
{
	if (global_cmuview_file) {
		fprintf(fp,"%s",cmd);
		return 0;
	} else {
		return wg.SendCommand(cmd);
	}
}

int cmumovie_config(int _width, int _height)
{
	char buf[100];
	width=_width;
	height=_height;
	int mw,mh;
	mw=width*2+1;
	mh=height*2+1;
	sprintf(buf,"cmuconfig %i %i 0 0 %i %i\n",mw,mh,mw*10,mh*10);
	if (cmumovie_command(buf)==ERROR) return ERROR;

	return cmumovie_command("cmuruler 0.005 \"nanoFoobars\"\n");
}

static int draw_line(int x1, int y1, int x2, int y2, int l, const char *color)
{
	int mx1,mx2,my1,my2;
	mx1=x1; mx2=x2-1; my1=y1; my2=y2-1;
	if (l==1 || l==3) {
		mx1+=width+1;
		mx2+=width+1;
	}
	if (l==2 || l==3) {
		my1+=height+1;
		my2+=height+1;
	}
	
	char buf[100];
	if (mx1==mx2) {
		sprintf(buf,"rf %i.25 %i.25 %i.75 %i.75 \"%s\" ; \n",
			mx1,my1,mx2,my2,color);
	}
	else {
		sprintf(buf,"rf %i.25 %i.25 %i.75 %i.75 \"%s\" ; \n",
			mx1,my1,mx2,my2,color);
	}
	cmumovie_command(buf);
	return 0;
}
	

static int draw_filled_box(int x1, int y1, int x2, int y2, int l, const char *color)
{
	int mx1,mx2,my1,my2;
	mx1=x1; mx2=x2; my1=y1; my2=y2;
	if (l==1 || l==3) {
		mx1+=width+1;
		mx2+=width+1;
	}
	if (l==2 || l==3) {
		my1+=height+1;
		my2+=height+1;
	}

	char buf[100];
	sprintf(buf,"rf %i %i %i %i \"%s\" ; \n",mx1,my1, mx2,my2,color);
	cmumovie_command(buf);
	return 0;
}

static int drawbox(int x1, int y1, int x2, int y2, char *color)
{
	char buf[100];

	sprintf(buf,"ro %i %i %i %i \"%s\" ; \n",x1,y1,x2,y2,color);
	cmumovie_command(buf);

	return 0;
}

extern vector<net> vn;

int cmumovie_newframe_short()
{
	cmumovie_command("cmuframe\n");
	return 0;
}

int cmumovie_newframe()
{
	cmumovie_command("cmuframe\n");
	cmumovie_command("cmuclear\n");

	// Now we have to draw in the bounding box for each layer,
	// and all the pin locations
	drawbox(0,0,width,height,"RosyBrown"); // layer 0
	drawbox(width+1,0,width*2+1,height,"RosyBrown"); // layer 1
	drawbox(0,height+1,width,height*2+1,"RosyBrown"); // layer 2
	drawbox(width+1,height+1,width*2+1,height*2+1,"RosyBrown"); // layer 3

	int i,j;
	char buf[100];
	for (i=0;i<vn.size();i++) {
		for (j=0;j<vn[i].pins.size();j++) {
			// now we have a pin, draw it as a black box
			// on layer 0 and layer 1
			sprintf(buf,"rf %i %i %i %i \"black\" ;\n",
				vn[i].pins[j].x,vn[i].pins[j].y-1,
				vn[i].pins[j].x+1,vn[i].pins[j].y+2);
			cmumovie_command(buf);
			sprintf(buf,"rf %i %i %i %i \"black\" ;\n",
				vn[i].pins[j].x+width+1,vn[i].pins[j].y-1,
				vn[i].pins[j].x+width+2,vn[i].pins[j].y+2);
			cmumovie_command(buf);
		}
	}
	//	cmumovie_command("doZoomFIT\n");
	return 0;
}
	
int cmumovie_drawsourcedest(list<cell> &source, list<pin> &dest)
{
	list<cell>::iterator lci;
	for (lci=source.begin();lci!=source.end();++lci) {
		draw_filled_box(lci->x,lci->y,lci->x+1,lci->y+1,lci->l,
				"HotPink");
	}
	list<pin>::iterator lpi;
	for (lpi=dest.begin();lpi!=dest.end();++lpi) {
		draw_filled_box(lpi->x,lpi->y-1,lpi->x+1,lpi->y+2,0,"turquoise");
		draw_filled_box(lpi->x,lpi->y-1,lpi->x+1,lpi->y+2,1,"turquoise");
	}
}

int cmumovie_drawexpansion(list<cell> &expanded_cells)
{
	list<cell>::iterator lci;
	for (lci=expanded_cells.begin();lci!=expanded_cells.end();++lci) {
		draw_filled_box(lci->x,lci->y,lci->x+1,lci->y+1,lci->l,
				"SandyBrown");
	}
	return 0;
}

int cmumovie_drawnets(vector<list<wire_segment> > & vlws, vector<int> &vio)
{
	int i,j;
	//	fprintf(stderr,"cmumovie_drawnets: size=%i\n",vlc.size());
	for (i=0;i<vlws.size();i++) {
		vector<cell> vias;
		//list<wire_segment> lws=path_to_segment(vlc[i]);
		// now go and draw each wire segment
		list<wire_segment>::iterator lwsi;
		for (lwsi=vlws[i].begin();lwsi!=vlws[i].end();++lwsi) {
			if (lwsi->z1 != lwsi->z2) {
				cell c;
				c.x=lwsi->x1; c.y=lwsi->y1; c.l=lwsi->z1;
				vias.push_back(c);
				c.l=lwsi->z2;
				vias.push_back(c);
			} else {
				// WIRE
				int minx=min(lwsi->x1,lwsi->x2);
				int maxx=max(lwsi->x1,lwsi->x2);
				int miny=min(lwsi->y1,lwsi->y2);
				int maxy=max(lwsi->y1,lwsi->y2);
				//draw_filled_box(minx,miny,maxx+1,maxy+1,
				//	lwsi->z1,(vio[i]>0) ? "red" : "SpringGreen");
				draw_line(minx,miny,maxx+1,maxy+1,
					lwsi->z1,(vio[i]>0) ? "red" : "SpringGreen");


			}
		}
		for (j=0;j<vias.size();j++) {
			// VIA
			// we should draw a blue rect on both layers
			draw_filled_box(vias[j].x,vias[j].y,
					vias[j].x+1,vias[j].y+1,
					vias[j].l,"blue2");
		}
	}
	return 0;
}
